/*
 * fs_cli.c - command-line interpreter for file server
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "aun.h"
#include "fs_proto.h"
#include "fs_errors.h"
#include "fileserver.h"
#include "extern.h"

typedef void fs_cmd_impl __P((struct fs_context *, char *));

struct fs_cmd {
	char	*full;
	char	*min;
	fs_cmd_impl	*impl;
};

static fs_cmd_impl fs_cmd_i_am;
static fs_cmd_impl fs_cmd_lib;
static fs_cmd_impl fs_cmd_sdisc;

static int fs_cli_match __P((char *word, const struct fs_cmd *cmd));

static const struct fs_cmd cmd_tab[3] = {
	{"I", 		"I",	fs_cmd_i_am,	}, /* Odd case */
	{"LIB",		"LIB",	fs_cmd_lib,	},
	{"SDISC",      	"SDIS",	fs_cmd_sdisc,	},
};

#define NCMDS (sizeof(cmd_tab) / sizeof(cmd_tab[0]))

void fs_cli __P((struct fs_context *));

/*
 * Handle a command-line packet from a client.  This is rarely used by
 * new clients (but NetFS still uses *I am).
 */
void
fs_cli(c)
	struct fs_context *c;
{
	int i;
	char *head, *tail;
	c->req->data[strcspn(c->req->data, "\r")] = '\0';
	if (debug) printf("cli: [%s]", c->req->data);

	tail = c->req->data;
	if (*tail == '*') tail++;
	head = fs_cli_getarg(&tail);
	for (i = 0; i < NCMDS; i++) {
		if (fs_cli_match(head, &(cmd_tab[i]))) {
			(cmd_tab[i].impl)(c, tail);
			break;
		}
	}
	if (i == NCMDS) 
		fs_unrec(c);
}

/*
 * Work out if word is an acceptable abbreviation for cmd.  Mutilates
 * word in the process.
 */

static int
fs_cli_match(word, cmd)
	char *word;
	const struct fs_cmd *cmd;
{
	char *p;
	for (p = word; *p!='\0'; p++)
		*p = toupper(*p);
	p--;
	if (*p == '.') {
		/* Abbreviated command */
		*p = '\0';
		if (strstr(cmd->full, word) == cmd->full &&
		    strstr(word, cmd->min) == word)
			return 1;
	} else {
		if (strcmp(word, cmd->full) == 0)
			return 1;
	}
	return 0;
}

/*
 * A bit like strsep, only different.  Breaks off the first word
 * (vaguely defined) of the input.  Afterwards, returns a pointer to
 * the first word (null-terminated) and points stringp at the start of
 * the tail.  Destroys the input in the process.
 */

char *
fs_cli_getarg(stringp)
	char **stringp;
{
	char *start;
	/* Skip leading whitespace */
	for (; **stringp == ' '; (*stringp)++);
	switch (**stringp) {
	case '"':
   		/* Quoted string. */
                /*
		 * XXX There seems to be no way to embed double quotes
		 * in a quoted string (or at least, NetFiler doesn't
		 * know how).  For now, assume the first '"' ends the
		 * string.
		 */
		(*stringp)++;
		start = *stringp;
		*stringp = strchr(start, '"');
		if (*stringp == NULL)
			/* Badness -- unterminated quoted string. */
			*stringp = strchr(start, '\0');
		else {
			**stringp = '\0';
			(*stringp)++;
		}
		break;
	case '\0':
		/* End of string hit.  Return two null strings. */
		start = *stringp;
		break;
	default:
		/* Unquoted.  Terminates at next space or end. */
		start = *stringp;
		*stringp = strchr(start, ' ');
		if (*stringp == NULL)
			*stringp = strchr(start, '\0');
		else {
			**stringp = '\0';
			(*stringp)++;
		}
	}
	return start;
}

static void
fs_cmd_i_am(c, tail)
	struct fs_context *c;
	char *tail;
{
	/* FIXME user authentication would be nice */
	struct ec_fs_reply_logon reply;
	char *login, *password;
	if (strcasecmp(fs_cli_getarg(&tail), "am")) {
		fs_unrec(c);
		return;
	}
	login = fs_cli_getarg(&tail);
	password = fs_cli_getarg(&tail);
	if (debug) printf(" -> log on [%s:%s]\n", login, password);
#if 0
	/* Silly test authentication */
	if (strcmp(login, "bjh21")) {
		fs_err(c, 0xbc);
		return;
	}
	if (strcmp(password, "secret")) {
		fs_err(c, 0xbb);
		return;
	}
#endif
	/*
	 * They're authenticated, so add them to the list of clients.
	 * First, we see if this client's already logged on, and if
	 * so, log them off first.
	 */
	if (c->client)
		fs_delete_client(c->client);
	c->client = fs_new_client(c->from);
	if (c->client == NULL) {
		fs_error(c, 0xff, "Internal server error");
		return;
	}
	c->client->login = strdup(login);
	reply.std_tx.command_code = EC_FS_CC_LOGON;
	reply.std_tx.return_code = EC_FS_RC_OK;
	/* Initial user environment.  Note that we can't use the same handle twice. */
	reply.urd = fs_open_handle(c->client, urd);
	reply.csd = fs_open_handle(c->client, urd);
	reply.lib = fs_open_handle(c->client, "/");
	reply.opt4 = opt4;
	fs_reply(c, &(reply.std_tx), sizeof(reply));
}

static void
fs_cmd_sdisc(c, tail)
	struct fs_context *c;
	char *tail;
{
	struct ec_fs_reply_sdisc reply;

	if (debug) printf(" -> sdisc\n");
	if (c->client == NULL) {
		fs_error(c, 0xff, "Who are you?");
		return;
	}
	reply.std_tx.command_code = EC_FS_CC_LOGON;
	reply.std_tx.return_code = EC_FS_RC_OK;
	/* Reset user environment.  Note that we can't use the same handle twice. */
	fs_close_handle(c->client, c->req->urd);
	fs_close_handle(c->client, c->req->csd);
	fs_close_handle(c->client, c->req->lib);
	reply.urd = fs_open_handle(c->client, "/home/bjh21");
	reply.csd = fs_open_handle(c->client, "/home/bjh21");
	reply.lib = fs_open_handle(c->client, "/");
	fs_reply(c, &(reply.std_tx), sizeof(reply));
}

static void
fs_cmd_lib(c, tail)
	struct fs_context *c;
	char *tail;
{
	char *upath;
	struct stat st;
	struct ec_fs_reply_dir reply;

	upath = fs_unixify_path(c, fs_cli_getarg(&tail)); /* Free it! */
	if (fs_stat(upath, &st) == -1) {
		fs_errno(c);
		goto burn;
	}
	fs_close_handle(c->client, c->req->lib);
	reply.new_handle = fs_open_handle(c->client, upath);
	if (reply.new_handle == 0) {
		fs_err(c, EC_FS_E_MANYOPEN);
		goto burn;
	}
	reply.std_tx.command_code = EC_FS_CC_LIB;
	reply.std_tx.return_code = EC_FS_RC_OK;
	fs_reply(c, &(reply.std_tx), sizeof(reply));
burn:
	free(upath);
}
